#include <stdio.h>

FILE *infile;
FILE *outfile;

typedef struct t_MBData		/*	Macroblock data	*/
{
	long int Byte;
	unsigned int Bit;
	int dct_dc_y;
	int dct_dc_cb;
	int dct_dc_cr;
	int quant;
} t_MBData;

int get_dcs_y(void);
int get_dcs_c(void);
void put_dcs_y(int);
void put_dcs_c(int);
int vlc(int);
unsigned short swap16(unsigned short x);
unsigned int swap32(unsigned int x);

int main(int argc,char *argv[])
{ 
	int sizex;
	int sizey;
	int frames;
	int frame;
	int dct_dc_y;
	int dct_dc_cb;
	int dct_dc_cr;
	int quant;
	int flag;
	int mb;
	int mb_source;
	int intraquant;
	int block;
	int eob;
	int size;
	int diff;
	int absval;
	long int FrameByte;
	unsigned int FrameBit;
	unsigned int length;
	
	t_MBData * MBData;
		
	if (argc<3){
		printf("\nConverts an Singstar IPU-movie into an MPEG-Video\n20080103 - hawkear@gmx.de\n\n"
			"Usage:     %s <INFILE> <OUTFILE>\n\n"
			"Example:   %s movie.ipu myvideo.m2v\n\n",argv[0],argv[0]);
		exit(0);
	}

	if ((infile = fopen (argv[1], "rb")) == NULL)
	{
		fprintf (stderr,"Can't open %s\n",argv[1]);
		exit(1);
	}
	if (!(outfile=fopen(argv[2],"wb")))
	{
		fprintf(stderr,"Can't write to %s\n",argv[2]);
		exit(1);
	}

	if (0x6970756d != Get(32)) {
		fprintf (stderr,"%s is no IPU\n",argv[1]);
		exit(1);
	}
	putbits(0x6970756d,32);
	
	Get(32);		// Filesize
	putbits(0,32);	// Filesize Dummy
	
	sizex=Get(8);
	sizex=sizex+Get(8)*(1<<8);
	putbits(swap16(sizex),16);
	
	sizey=Get(8);
	sizey=sizey+Get(8)*(1<<8);
	putbits(swap16(sizey),16);

	frames=Get(8);
	frames=frames+Get(8)*(1<<8);
	frames=frames+Get(8)*(1<<16);
	frames=frames+Get(8)*(1<<24);
	putbits(swap32(frames),32);

	printf("%dx%d\n",sizex,sizey);
	printf("%02d:%02d:%02d.%02d\n\n",frames/25/60/60,(frames%(25*60*60))/25/60,(frames%(25*60))/25,frames%25);


	MBData = (t_MBData*) malloc (((sizex/16)*(sizey/16)+1)*sizeof(t_MBData));
	if (MBData==NULL) {
		fprintf (stderr,"Could not allocate memory\n");
		exit(1);
	}

	for (frame=0;frame<frames;frame++){
		printf("Frame: %d/%d\r",frame+1,frames);
		flag = Get(8);
		putbits(flag,8);
		
		if (flag & 32) {
			fprintf (stderr,"Intra VLC format not supported\n");
			exit(1);
		}

		// Get Macroblock-infos
		dct_dc_y = 0;
		dct_dc_cb = 0;
		dct_dc_cr = 0;
		quant = 1;
		for(mb=0;mb<(sizex/16)*(sizey/16);mb++) {
			if (mb>0) {
				if (!Get(1)) {
					fprintf (stderr,"MBA_Incr wrong in IPU\n");
					exit(1);
				}
			}
			// Save position (in IPU) of Macroblock 
			getpos(&MBData[mb].Byte,&MBData[mb].Bit);

			if (Get(1)) {
				intraquant = 0;
			} else {
				if (!Get(1)) {
					fprintf (stderr,"MBT wrong in IPU\n");
					exit(1);
				}
				intraquant = 1;
			}

			if (flag & 4) {
				Get(1);
			}

			if (intraquant) {
				quant = Get(5);
			}
			MBData[mb].quant = quant;

			for(block=0;block<6;block++) {
				if (block<4) {
					if (size = get_dcs_y()) {
						diff = Get(size);
						if (!(diff & (1 << (size - 1))))
							diff = (-1 << size) | (diff + 1);
						dct_dc_y += diff; 
					}
					if (block==0) 
						MBData[mb].dct_dc_y = dct_dc_y;
				} else {
					if (size = get_dcs_c()) {
						diff = Get(size);
						if (!(diff & (1 << (size - 1))))
							diff = (-1 << size) | (diff + 1);
					}
					if (block==4) {
						if (size)
							dct_dc_cb += diff;
						MBData[mb].dct_dc_cb = dct_dc_cb;
					} else {
						if (size)
							dct_dc_cr += diff;
						MBData[mb].dct_dc_cr = dct_dc_cr;
					}
				}
				do {
					eob = vlc(0);
					if (eob == 0) Get(1);
					} while (eob!=1);
			}
		}

		getpos(&FrameByte,&FrameBit);	// Position speichern
		
		// Write Macroblocks
		dct_dc_y = 0;
		dct_dc_cb = 0;
		dct_dc_cr = 0;
		quant = 1;
		for(mb=0;mb<(sizex/16)*(sizey/16);mb++) {
//			mb_source = (mb % (sizex / 16)) * (sizey / 16)+ mb / (sizex / 16);	// Singstar to normal
//			mb_source = mb;														// normal to normal
			mb_source = (mb % (sizey / 16)) * (sizex / 16)+ mb / (sizey / 16);	// normal to Singstar

			if (mb>0) {
				putbits(1,1);
			}
			
			setpos(MBData[mb_source].Byte,MBData[mb_source].Bit);

			if (Get(1)) {
				intraquant = 0;
			} else {
				if (!Get(1)) {
					fprintf (stderr,"MBT wrong in IPU\n");
					exit(1);
				}
				intraquant = 1;
			}

			if (mb == 0 || (MBData[mb_source].quant != quant))
				putbits(1,2);
			else
				putbits(1,1);
				
			
			if (flag & 4) {
				putbits(Get(1),1);
			}

			if (intraquant) {
				quant = Get(5);
			}
			if (mb == 0 || (MBData[mb_source].quant != quant)) {
				putbits(MBData[mb_source].quant,5);
				quant = MBData[mb_source].quant;
			}

			for(block=0;block<6;block++) {
				if (block==0) {
					Get(get_dcs_y());	// DCT_DC in Eingabestream berspringen
					diff = MBData[mb_source].dct_dc_y - dct_dc_y;
					dct_dc_y = MBData[mb_source].dct_dc_y;

					absval = (diff<0) ? -diff : diff; /* abs(val) */
					size = 0;
					while (absval) {
						absval >>= 1;
						size++;
					}
					put_dcs_y(size);
					absval = diff;
					if (absval<=0)
						absval += (1<<size) -1;
					putbits(absval,size);
				} else if (block>3)	{
					Get(get_dcs_c());	// DCT_DC in Eingabestream berspringen
					if (block==4) {
						diff = MBData[mb_source].dct_dc_cb - dct_dc_cb;
						dct_dc_cb = MBData[mb_source].dct_dc_cb;
					} else {
						diff = MBData[mb_source].dct_dc_cr - dct_dc_cr;
						dct_dc_cr = MBData[mb_source].dct_dc_cr;
					}
					absval = (diff<0) ? -diff : diff; /* abs(val) */
					size = 0;
					while (absval) {
						absval >>= 1;
						size++;
					}
					put_dcs_c(size);
					absval = diff;
					if (absval<=0)
						absval += (1<<size) -1;
					putbits(absval,size);
				} else {
					size = get_dcs_y();
					put_dcs_y(size);
					diff = Get(size);
					putbits(diff,size);		// DCT_DC kopieren (Blcke 1,2,3)
					if (size) {
						if (!(diff&(1<<(size - 1))))
							diff = (-1 << size) | (diff + 1);
						dct_dc_y += diff;
					}
				}
				do {
					eob = vlc(1);
					if (eob == 0)
						putbits(Get(1),1);
				} while (eob!=1);
			}
			

			
			
		}
		putbuf();
	
		setpos(FrameByte,FrameBit);
		
		// Jump to End of Frame
		if (!Next_Start_Code()) {
			fprintf (stderr,"End of Stream\n");
			exit(1);
		}

		if(Get(32) != 0x000001b0) {
			fprintf (stderr,"No 1b0\n");
			exit(1);
		}
		putbits(0x000001b0,32);
	}
	

	putbits(0x1b1,32);		// Ende
  
	free (MBData);
	fclose(infile);   
	putbuf();	// Write Buffer, if not empty
	
	fseeko64(outfile,0,SEEK_END);
	length = (unsigned int)ftello64(outfile);
	fseeko64(outfile,4,SEEK_SET);
	putbits(swap32(length-8),32);

	fclose(outfile);   
}


int vlc(int write){
	int bits;
	int level=0;
	
	bits = Get(2);
	if (write) putbits(bits,2);
	if(bits ==	2)
		return(1);		// 10 - EOB
	if(bits ==	3)
		return(0);		// 11
	if(bits ==	1) {
		bits = Get(1);
		if (write) putbits(bits,1);
		if (bits)
			return(0);
		else {
			bits = Get(1);
			if (write) putbits(bits,1);
			return(0);
		}
	}
	// 00
	bits = Get(1);
	if (write) putbits(bits,1);
	if (bits) {
		// 001
		bits = Get(2);
		if (write) putbits(bits,2);
		if (bits < 1) {
			// 00100
			bits = Get(3);
			if (write) putbits(bits,3);
		}
		// 001xx
		return(0);
	} else {
	// 000
		bits = Get(3);
		if (write) putbits(bits,3);
		if(bits >=	4)
			return(0);
		if(bits >=	2) {
			bits = Get(1);
			if (write) putbits(bits,1);
			return(0);
		}
		if(bits) {
			bits = Get(18);
			if (write) putbits(bits,18);
			return(2);	// Escape
		}
		bits = Get(1);
		if (write) putbits(bits,1);
		if(bits) {
			bits = Get(3);
			if (write) putbits(bits,3);
			return(0);
		}
		do {
			bits = Get(1);
			if (write) putbits(bits,1);
			level++;
		} while (!bits && level<6);
		if (level<6) {
			bits = Get(4);
			if (write) putbits(bits,4);
			return(0);
		} else {
			fprintf (stderr,"Invalid VLC\n");
			exit(1);
		}
	}
}
int get_dcs_y(void){
	int bits;
	
	if(!(bits = Get(2)))
		return(1);		// 00
	if(bits ==	1)
		return(2);		// 01
	bits <<= 1;
    bits |= Get(1);
	if(bits ==	4)
		return(0);		// 100
	if(bits ==	5)
		return(3);		// 101
	if(bits ==	6)
		return(4);		// 110
	if(!Get(1))
		return(5);
	if(!Get(1))
		return(6);
	if(!Get(1))
		return(7);
	if(!Get(1))
		return(8);
	if(!Get(1))
		return(9);
	if(!Get(1))
		return(10);
	return(11);
}
int get_dcs_c(void){
	int bits;

	if(!(bits = Get(2)))
		return(0);		// 00
	if(bits ==	1)
		return(1);		// 01
	if(bits ==	2)
		return(2);		// 10
	if(!Get(1))
		return(3);
	if(!Get(1))
		return(4);
	if(!Get(1))
		return(5);
	if(!Get(1))
		return(6);
	if(!Get(1))
		return(7);
	if(!Get(1))
		return(8);
	if(!Get(1))
		return(9);
	if(!Get(1))
		return(10);
	return(11);
}
void put_dcs_y(int len){
	switch (len) {
		case 0:
			putbits(4,3);
			break;
		case 1:
			putbits(0,2);
			break;
		case 2:
			putbits(1,2);
			break;
		case 3:
			putbits(5,3);
			break;
		case 4:
			putbits(6,3);
			break;
		case 5:
			putbits(14,4);
			break;
		case 6:
			putbits(30,5);
			break;
		case 7:
			putbits(62,6);
			break;
		case 8:
			putbits(126,7);
			break;
		case 9:
			putbits(254,8);
			break;
		case 10:
			putbits(510,9);
			break;
		case 11:
			putbits(511,9);
			break;
	}
}
void put_dcs_c(int len){
	switch (len) {
		case 0:
			putbits(0,2);
			break;
		case 1:
			putbits(1,2);
			break;
		case 2:
			putbits(2,2);
			break;
		case 3:
			putbits(6,3);
			break;
		case 4:
			putbits(14,4);
			break;
		case 5:
			putbits(30,5);
			break;
		case 6:
			putbits(62,6);
			break;
		case 7:
			putbits(126,7);
			break;
		case 8:
			putbits(254,8);
			break;
		case 9:
			putbits(510,9);
			break;
		case 10:
			putbits(1022,10);
			break;
		case 11:
			putbits(1023,10);
			break;
	}
}

unsigned short swap16(unsigned short x){
    x = (x>>8) | 
        (x<<8);
	return x;
}

unsigned int swap32(unsigned int x){
    x = (x>>24) | 
        ((x<<8) & 0x00FF0000) |
        ((x>>8) & 0x0000FF00) |
        (x<<24);
	return x;
}
